//	Image Noise Removal Filter
//	Copyright (C) 2002 Karlis Freivalds
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//---------------------------------------------------------------------------
#include <vcl.h>
#include <math.h>
#pragma hdrstop

#include "Unit1.h"
#include "about.h"
#include "Parameters.h"
#include "jpeg.hpp"

#include <stdio.h>
#include "yiq.h"
#include "median.h"
#include "colorDenoise.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;

Graphics::TBitmap *Bitmap = NULL;
Graphics::TBitmap *viewBitmap = NULL;
Graphics::TBitmap *undoBitmap = NULL;
int width,height;

int distWeight = 12*256;
int radius = 2;

//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------

void processColor(float radius, int w, float radiusCr, bool reduceCr,
    float radiusCb)
{

  width = Bitmap->Width;
  height = Bitmap->Height;
  int size = width*height;
  int *Y = new int[size];
  int *I = new int[size];
  int *Q = new int[size];

  RGB2YIQ(Bitmap, Y, I, Q);

  colorDenoise noiser;
  noiser.setParameters(radius, radiusCr, radiusCb,w,reduceCr);
  noiser.process(Y, Q, I,width,height);

  YIQ2RGB(Bitmap, Y, I, Q);

  delete [] Y;
  delete [] I;
  delete [] Q;
}

void prepareChroma(int luma)
{
    if(!viewBitmap)
    {
        viewBitmap = new Graphics::TBitmap();
        viewBitmap->Assign(Bitmap);
    }

  width = Bitmap->Width;
  height = Bitmap->Height;
  int size = width*height;
  int *Y = new int[size];
  int *I = new int[size];
  int *Q = new int[size];

  RGB2YIQ(Bitmap, Y, I, Q);

    int	h = Bitmap->Height;
    int *data;
    if(luma==1)
        data= Y;
    else if(luma==2)
        data = I;
    else
        data = Q;

    do
    {
        int	w = Bitmap->Width;
        do
        {
            int a = (*data++)/256;

            if(luma)
            {
                if(a<0) a = 0;
                if(a>255) a = 255;
                viewBitmap->Canvas->Pixels[width-w][height-h] = RGB(a,a,a);
            }
            else
            {
                int c = Bitmap->Canvas->Pixels[width-w][height-h];
                int r = (c &255)-a+128;
                int g = ((c>>8) &255)-a+128;
                int b = ((c>>16) &255)-a+128;
                if(r<0) r=0;
                if(g<0) g=0;
                if(b<0) b=0;
                if(r>255) r = 255;
                if(g>255) g = 255;
                if(b>255) b = 255;

                viewBitmap->Canvas->Pixels[width-w][height-h] = RGB(r,g,b);
            }
        } while(--w);

    } while(--h);


  delete [] Y;
  delete [] I;
  delete [] Q;
}

void prepareNoise()
{
    if(!Bitmap || !undoBitmap)
        return;

    if(!viewBitmap)
    {
        viewBitmap = new Graphics::TBitmap();
        viewBitmap->Assign(Bitmap);
    }

  width = Bitmap->Width;
  height = Bitmap->Height;

  int h = Bitmap->Height;

    do
    {
        int	w = Bitmap->Width;
        do
        {
                int c = Bitmap->Canvas->Pixels[width-w][height-h];
                int c1 = undoBitmap->Canvas->Pixels[width-w][height-h];

                int r = ((c &255)-(c1 &255));
                int g = ((c>>8) &255)-((c1>>8) &255);
                int b = ((c>>16) &255)-((c1>>16) &255);
                r = r*2+128;
                g = g*2+128;
                b = b*2+128;

                if(r<0) r=0;
                if(g<0) g=0;
                if(b<0) b=0;
                if(r>255) r = 255;
                if(g>255) g = 255;
                if(b>255) b = 255;

                viewBitmap->Canvas->Pixels[width-w][height-h] = RGB(r,g,b);
        } while(--w);

    } while(--h);
}

void __fastcall TForm1::FormPaint(TObject *Sender)
{
  if(viewBitmap)
  {
      Canvas->Draw(0,0,viewBitmap);
  }
}
//---------------------------------------------------------------------------
void __fastcall TForm1::Open1Click(TObject *Sender)
{
    if(OpenDialog1->Execute())
    {
        Bitmap = new Graphics::TBitmap();
        try
        {
            Bitmap->LoadFromFile(OpenDialog1->FileName);
        }catch(Exception *e)
        {
            //Bitmap = NULL;
            try
            {
                TFileStream *stream = new TFileStream(OpenDialog1->FileName,fmOpenRead);
                TJPEGImage *jp = new TJPEGImage;
                jp->LoadFromStream(stream);
                Bitmap->Assign(jp);
                delete jp;
                delete stream;
            }catch(Exception *e)
            {
                delete Bitmap;
                //delete jp;
                //delete stream;
                Bitmap = NULL;
            }
        }
        viewBitmap = new Graphics::TBitmap();
        viewBitmap->Assign(Bitmap);
        Image1->Checked = true;
        Repaint();
    }
}
//---------------------------------------------------------------------------



void __fastcall TForm1::Save1Click(TObject *Sender)
{
    if(Bitmap && SaveDialog1->Execute())
    {
        TFileStream *stream = new TFileStream(SaveDialog1->FileName,fmCreate);

        if(SaveDialog1->FilterIndex==1)
        {
            Bitmap->SaveToStream(stream);
        }else if(SaveDialog1->FilterIndex==2)
        {
            TJPEGImage *jp = new TJPEGImage;
            jp->Assign(Bitmap);
            jp->SaveToStream(stream);

            delete jp;
        }

        delete stream;
    }

}
//---------------------------------------------------------------------------

void __fastcall TForm1::Denoise1Click(TObject *Sender)
{
    if(Bitmap)
    {
        if(Form2->ShowModal()==mrOk)
        {
            if(!undoBitmap)
                undoBitmap = new Graphics::TBitmap;
            undoBitmap->Assign(Bitmap);

            processColor(Form2->radius, Form2->distWeight,
                Form2->radiusCr,
                Form2->reductionCr,
                Form2->radiusCb);

            Image1->Checked = true;
            viewBitmap->Assign(Bitmap);
            Repaint();
        }
    }
}
//---------------------------------------------------------------------------
void __fastcall TForm1::Image1Click(TObject *Sender)
{
    if(!Bitmap) return;
    Image1->Checked = true;
    viewBitmap->Assign(Bitmap);
    Repaint();
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Luma1Click(TObject *Sender)
{
    if(!Bitmap) return;
    Luma1->Checked = true;
    prepareChroma(true);
    Repaint();
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Chroma1Click(TObject *Sender)
{
    if(!Bitmap) return;
//    Chroma1->Checked = true;
    prepareChroma(false);
    Repaint();
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Removednoise1Click(TObject *Sender)
{
    if(!Bitmap) return;
    Removednoise1->Checked = true;
    prepareNoise();
    Repaint();
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Undo1Click(TObject *Sender)
{
    if(undoBitmap)
    {
        Bitmap->Assign(undoBitmap);
        viewBitmap->Assign(undoBitmap);
        Image1->Checked = true;
        delete undoBitmap;
        undoBitmap = NULL;
    }

    Repaint();
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Cr1Click(TObject *Sender)
{
    if(!Bitmap) return;
    Cr1->Checked = true;
    prepareChroma(3);
    Repaint();
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Cb1Click(TObject *Sender)
{
    if(!Bitmap) return;
    Cb1->Checked = true;
    prepareChroma(2);
    Repaint();

}
//---------------------------------------------------------------------------

void __fastcall TForm1::About1Click(TObject *Sender)
{
    Form3->ShowModal();
}
//---------------------------------------------------------------------------
void applyMedian(bool strong)
{
  width = Bitmap->Width;
  height = Bitmap->Height;
  int size = width*height;
  int *Y = new int[size];
  int *I = new int[size];
  int *Q = new int[size];

  RGB2YIQ(Bitmap, Y, I, Q);

  if(!strong)
      medianFilter(Y, Q, I, width, height);
  else
      strongMedianFilter(Y, Q, I, width, height);


  YIQ2RGB(Bitmap, Y, I, Q);

  delete [] Y;
  delete [] I;
  delete [] Q;
}

void __fastcall TForm1::Median1Click(TObject *Sender)
{
    if(Bitmap)
    {
            if(!undoBitmap)
                undoBitmap = new Graphics::TBitmap;
            undoBitmap->Assign(Bitmap);

            applyMedian(false);

            Image1->Checked = true;
            viewBitmap->Assign(Bitmap);
            Repaint();
    }

}
//---------------------------------------------------------------------------


void __fastcall TForm1::Strongmedian1Click(TObject *Sender)
{
    if(Bitmap)
    {
            if(!undoBitmap)
                undoBitmap = new Graphics::TBitmap;
            undoBitmap->Assign(Bitmap);

            applyMedian(true);

            Image1->Checked = true;
            viewBitmap->Assign(Bitmap);
            Repaint();
    }
}
//---------------------------------------------------------------------------

