//	Image Noise Removal Filter
//	Copyright (C) 2002 Karlis Freivalds
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <stdlib.h>
#include <string.h>

const deltaY = 15;
const deltaCr = 30;
const deltaCb = 100;

#define MINMAX(A, B, Mi, Ma) if((A) < (B)) {Mi = (A); Ma = (B);} else {Mi = (B); Ma = (A);}

#define MEDIAN3(A,B,C,Lo,Med,Hi) \
{\
	int t1, t2, t3;\
	MINMAX(B, C, t1, t2)\
	MINMAX(A, t1, Lo, t3)\
	MINMAX(t3, t2, Med, Hi)\
}

int median9(int *data)
{
	int a8, a9, a10, a11, a13, a14, a15, a16, a17; 
	int a18, a19, a23, a24, a28;	
	int result;

	MEDIAN3(data[0],data[1], data[2], a8, a9, a10);
	MEDIAN3(data[3],data[4], data[5], a11, a13, a14);
	MINMAX(a8, a11, int nu, a18);
	MEDIAN3(data[6],data[7], data[8], a15, a16, a17);
	MINMAX(a18, a15, int nu, a23);
	MINMAX(a14, a17, a19, int nu);
	MINMAX(a10, a19, a24, int nu);
	MEDIAN3(a9,a13, a16, int nu, a28, int nu);
	MEDIAN3(a23,a28, a24, int nu, result, int nu);

	return result;
}

bool isEdgeY(int *Y, int pitch, int delta)
{
	int thr = delta*256;
	int c = *Y;

	int c1 = *(Y+1);
    if(abs(c-c1) > thr) return true;

	int c2 = *(Y-1);
    if(abs(c-c2) > thr) return true;

	int c3 = *(Y-pitch);
    if(abs(c-c3) > thr) return true;

	int c4 = *(Y+pitch);
    if(abs(c-c4) > thr) return true;

    int maxValue = max(max(max(c1,c2),c3),c4);
    if(c >= maxValue+thr/2) return true;
    int minValue = min(min(min(c1,c2),c3),c4);
    if(c <= minValue-thr/2) return true;

    return false;
}

void median(int *Y, int *Cr, int *Cb, int pitch, int &resY, int &resR, int &resB)
{
    if(isEdgeY(Y, pitch,deltaY) || isEdgeY(Cr, pitch,deltaCr) ||
        isEdgeY(Cb, pitch,deltaCb))
    {
        resY = *Y;
        resR = *Cr;
        resB = *Cb;
        return;
    }

    int y[9],cr[9],cb[9];
    int counter = 0;

    for(int k=-1;k<=1;k++)
      for(int i=-1;i<=1;i++)
      {

		  int adr = k*pitch + i;
          y[counter] = Y[adr];
          cr[counter] = Cr[adr];
          cb[counter] = Cb[adr];
          counter++;
      }

    resY = median9(y);
    resR = median9(cr);
    resB = median9(cb);
}

void strongMedianFilter(int *Y, int *Cr, int *Cb, int width, int height)
{
	int w;
	int *Y1 = Y;
	int *I1 = Cr;
	int *Q1 = Cb;

    int h = height;
    do {
        w = width;

		do
		{
			if(h < height && h>1 && w >1 && w < width)
			{
                int y, r, b;
				median(Y1, I1, Q1, width,y,r,b);
                *Y1 = y;
                *I1 = r;
                *Q1 = b;
			}

			Y1++;
			I1++;
			Q1++;
		} while(--w);
    } while(--h);

}

void medianFilter(int *Y, int *Cr, int *Cb, int width, int height)
{
	int w;
    int size = width*height;
	int *Y1 = new int[size];
    int *I1 = new int[size];
    int *Q1 = new int[size];

    memcpy(Y1,Y,size*sizeof(int));
    memcpy(I1,Cr,size*sizeof(int));
    memcpy(Q1,Cb,size*sizeof(int));

    int counter = 0;

    int h = height;
    do {
        w = width;

		do
		{
			if(h < height && h>1 && w >1 && w < width)
			{
                int y, r, b;
				median(Y1+counter, I1+counter, Q1+counter, width,y,r,b);
                Y[counter] = y;
                Cr[counter] = r;
                Cb[counter] = b;
			}
            counter++;
		} while(--w);
    } while(--h);

    delete [] Y1;
    delete [] I1;
    delete [] Q1;


}


