//	Image Noise Removal Filter
//	Copyright (C) 2002 Karlis Freivalds
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <stdlib.h>
#include <string.h>
#include "queue.h"

PriorityQueue::PriorityQueue(int count,int maxrange_){
    maxsize=count;
    last=0;
    queue=(QueueElement *)malloc(count*sizeof(QueueElement));
    maxrange=maxrange_;
    places=(int *)malloc(maxrange*sizeof(int));
    memset(places,-1,maxrange*sizeof(int));
}

PriorityQueue::PriorityQueue(int count){
    maxsize=count;
    last=0;
    queue=(QueueElement *)malloc(count*sizeof(QueueElement));
    maxrange=count;
    places=(int *)malloc(maxrange*sizeof(int));
    memset(places,-1,maxrange*sizeof(int));
}

PriorityQueue::~PriorityQueue(){
    free(places);
    free(queue);
}


void PriorityQueue::AddElem(QueueElement e){

     int p=places[e.value];
     if(p>=0 && e.key>=queue[p].key) return;//alredy in with smaller key
     int i,j;

     if(p<0){
         if (last >= maxsize) return;//  error('queue is full');
         i = last++;
     }else i=p;

     queue[i]=e;

     while (i>0 && queue[i].key < queue[j=((i-1) >> 1)].key){
           // push x up the tree by exchanging it with its parent.
           QueueElement temp = queue[i];
           queue[i]          = queue[j];
           queue[j] = temp;
           places[queue[i].value]=i;
           i = j;  // update location of current element
     }
     places[queue[i].value]=i;

}

void PriorityQueue::GetElem(QueueElement *e){
      if(last == 0) return;// error('priority queue is empty.');
      *e = *queue;  //    we'll return the root
      places[e->value]=-1;
      if(!--last) return;
      *queue = queue[last];
      int i = 0;
      int j;
      while((j=2*i+2) <= last){
           if ((queue[j-1].key < queue[j].key)||
               (j == last)) //  only has one child
                j--;    //left child

           if (queue[i].key > queue[j].key){
                // swap parent and child
                QueueElement temp = queue[i];
                queue[i] = queue[j];
                queue[j] = temp;
                places[queue[i].value]=i;
                i = j;
           }
           else break; // don't push further down
      }
// pushed all the way to leaf node
      places[queue[i].value]=i;

}
