//	Image Noise Removal Filter
//	Copyright (C) 2002 Karlis Freivalds
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "reduce.h"

void shrinkLine2(int *indata, int *outData, int count)
{
    *outData++ = (*indata + indata[count])/2;
    indata++;

    for(int i=0;i<(count-1)/2;i++)
    {
        *outData++ = (indata[0]+indata[1]+indata[count]+indata[count+1])>>2;
        indata+=2;
    }

    if((count & 1)==0)
        *outData = (*indata + indata[count])/2;
}

void shrinkLine1(int *indata, int *outData, int count)
{
    *outData++ = *indata++;

    for(int i=0;i<(count-1)/2;i++)
    {
        *outData++ = (*indata+indata[1])>>1;
        indata+=2;
    }

    if((count & 1)==0)
        *outData = *indata;
}

void shrink2x(int *indata, int *outData, int w, int h)
{
    int newW = (w+2)/2;
    shrinkLine1(indata, outData, w);

    indata += w;
    outData+=newW;

    for(int i=0;i<(h-1)/2;i++)
    {
        shrinkLine2(indata, outData, w);
        indata += 2*w;
        outData += newW;
    }

    if((h & 1)==0)
        shrinkLine1(indata, outData, w);
}

void unShrinkLine1(int *indata, int *outData, int count)
{
    *outData++ = *indata++;

    for(int i=0;i<(count-1)/2;i++)
    {
        outData[0] = outData[1] = *indata++;
        outData+=2;
    }

    if((count & 1)==0)
        *outData = *indata;
}

void unShrinkLine2(int *indata, int *outData, int count)
{
    outData[0] = outData[count] = *indata++;
    outData++;
    int newW = (count+2)/2;
    int limit = (count-1)/2;

    if((count & 1)!=0)
        limit--;

    for(int i=0;i<limit;i++)
    {
        outData[0] = (indata[0]*9+(indata[-1]+indata[-newW])*3+indata[-newW-1])>>4;
        outData[1] = (indata[0]*9+(indata[+1]+indata[-newW])*3+indata[-newW+1])>>4;
        outData[count]= (indata[0]*9+(indata[-1]+indata[newW])*3+indata[newW-1])>>4;
        outData[count+1]= (indata[0]*9+(indata[1]+indata[newW])*3+indata[newW+1])>>4;
        //outData[count+1] = outData[count] = outData[0] = outData[1] = *indata++;
        indata++;
        outData+=2;
    }

    if((count & 1)!=0)
    {
        outData[0] = (indata[0]*9+(indata[-1]+indata[-newW])*3+indata[-newW-1])>>4;
        outData[1] = (indata[0]*3+indata[-newW])>>2;
        outData[count]= (indata[0]*9+(indata[-1]+indata[newW])*3+indata[newW-1])>>4;
        outData[count+1]= (indata[0]*3+indata[+newW])>>2;
        indata++;
        outData+=2;
    }

    if((count & 1)==0)
        outData[0] = outData[count] = *indata;
}

void unShrinkLine2Last(int *indata, int *outData, int count)
{
    outData[0] = outData[count] = *indata++;
    outData++;
    int newW = (count+2)/2;

    int limit = (count-1)/2;

    if((count & 1)!=0)
        limit--;

    for(int i=0;i<limit;i++)
    {
        outData[0] = (indata[0]*9+(indata[-1]+indata[-newW])*3+indata[-newW-1])>>4;
        outData[1] = (indata[0]*9+(indata[+1]+indata[-newW])*3+indata[-newW+1])>>4;
        outData[count]= (indata[0]*3+indata[-1])>>2;
        outData[count+1]= (indata[0]*3+indata[1])>>2;
        indata++;
        outData+=2;
    }

    if((count & 1)!=0)
    {
        outData[0] = (indata[0]*9+(indata[-1]+indata[-newW])*3+indata[-newW-1])>>4;
        outData[1] = (indata[0]*3+indata[-newW])>>2;
        outData[count]= (indata[0]*3+indata[-1])>>2;
        outData[count+1]= indata[0];
        indata++;
        outData+=2;
    }


    if((count & 1)==0)
        outData[0] = outData[count] = *indata;
}

void unShrink2x(int *outData, int *inData, int w, int h)
{
    int newW = (w+2)/2;
    unShrinkLine1(inData, outData, w);

    inData += newW;
    outData+=w;

    int newH = (h-1)/2;

    if((h & 1)!=0)
        newH--;

    for(int i=0;i<newH;i++)
    {
        unShrinkLine2(inData, outData, w);
        outData += 2*w;
        inData += newW;
    }

    if((h & 1)!=0)
    {
        unShrinkLine2Last(inData, outData, w);
        outData += 2*w;
        inData += newW;
    }

    if((h & 1)==0)
        unShrinkLine1(inData, outData, w);
}

reducedSmoothing::reducedSmoothing()
{
    this->reduce = false;
}

void reducedSmoothing::process(int *data, int width, int height)
{
    if(reduce)
    {
        newWidth = (width+2)/2;
        newHeight = (height+2)/2;

        int *newData = new int[newWidth*newHeight];
        shrink2x(data, newData, width, height);

        adaptiveSmoothing::process(newData, newWidth, newHeight);

        unShrink2x(data, newData, width, height);

        delete [] newData;
    }
    else
    {
        adaptiveSmoothing::process(data, width, height);
    }
}

void reducedSmoothing::setParameters(int radius, int preserveEdges,bool doReducing)
{
    this->reduce = doReducing;
    if(reduce) radius = max(radius/2,1);
    adaptiveSmoothing::setParameters(radius, preserveEdges);
}
