//	Image Noise Removal Filter
//	Copyright (C) 2002 Karlis Freivalds
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef smooth_H
#define smooth_H

#include "queue.h"

#define sqr(x) ((x)*(x))
#define max(a, b) ((a)>(b) ? (a) : (b))
#define min(a, b) ((a)<(b) ? (a) : (b))

struct point
{
    int x;
    int y;
    float distance;
};

/*
    This class performs filtering of one color plane
*/

class adaptiveSmoothing
{
public:
    void process(int *data, int width, int height);
    void setParameters(float radius, int preserveEdges);
    adaptiveSmoothing();
    void setEdgeGraph(int **graph);

private:
    void enhance();
    int enhancePixel(int x, int y);
    int search(int x, int y);
    void createFilter();
    void createDistMap();

    int width;
    int height;
    int range;
    int maxLength;
    point *hist;
    int *data;
    char *seen;
    PriorityQueue *Q;
    int distWeight;
    int maxDistance;
    double *filter;
    double invFilterLen;
    float *distMap;
    int **edgeGraph;
    float radius;
};

#endif
