//	Image Noise Removal Filter
//	Copyright (C) 2002 Karlis Freivalds
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <vcl.h>
#pragma hdrstop

//	Y	= 0.30R + 0.59G + 0.11B
//	R-Y	= 0.70R - 0.59G - 0.11B (179, -151, -28)
//	B-Y	=-0.30R - 0.59G + 0.89B (-77, -151, 228)
//
//	Inverse matrix:
//
//	R = Y + (R-Y)
//	G = Y - 0.50847(R-Y) - 0.18644(B-Y)
//	B = Y + (B-Y)


void RGB2YIQ(Graphics::TBitmap *Bitmap, int *Y, int *I, int *Q)
{
    int width = Bitmap->Width;
    int height = Bitmap->Height;
	int	h = Bitmap->Height;
    
    do {
        int	w = Bitmap->Width;
        do {
            int c = Bitmap->Canvas->Pixels[width-w][height-h];
            int b = (c>>16)&255;
            int g = (c>>8)&255;
            int r = c&255;

            int y = r*77 + g*151 + b*28;
            *Y++ = y;
            *Q++ = r*256 - y+128*256;
            *I++ = b*256 - y+128*256;

        } while(--w);

    } while(--h);
}

void YIQ2RGB(Graphics::TBitmap *Bitmap, int *Y, int *I, int *Q)
{
    int width = Bitmap->Width;
    int height = Bitmap->Height;
	int	h = Bitmap->Height;
		do {
			int	w = Bitmap->Width;
			do {

				int y = *Y++;
				int cr = (*Q++)-128*256;
				int cb = (*I++)-128*256;

				int r = (y + cr)/256;
				int g = (y - (130*cr + 48*cb)/256)/256;
				int b = (y + cb)/256;

                if(r<0) r = 0;
                if(g<0) g = 0;
                if(b<0) b = 0;
                if(r>255) r = 255;
                if(g>255) g = 255;
                if(b>255) b = 255;

				Bitmap->Canvas->Pixels[width-w][height-h] = RGB(r,g,b);
			} while(--w);

		} while(--h);
	}

